from pydantic import TypeAdapter

from .data import Data


DataModel = TypeAdapter(Data)


def from_json_schema(definition: dict) -> Data:
    definition = process_node(definition, definition)

    return DataModel.validate_python(definition)


def process_node(definition: dict, root: dict) -> dict:
    if "$ref" in definition:
        ref: str = definition.pop("$ref")
        # Try to dereference for now, in the future we may want to use
        # sdfData to store definitions
        fragments: list[str] = ref.split("/")
        assert fragments[0] == "#", "Only internal references supported"
        referenced = root
        for fragment in fragments[1:]:
            referenced = referenced[fragment]
        definition = {**referenced, **definition}

    if "title" in definition:
        # SDF uses label instead of title
        definition["label"] = definition.pop("title")

    if "anyOf" in definition:
        definition = convert_anyof(definition["anyOf"], root)
    else:
        # Can't be null
        definition["nullable"] = False

    if "enum" in definition:
        # Could maybe be replaced with sdfChoice
        definition = convert_enum(definition)

    if definition.get("format") == "binary":
        definition["sdfType"] = "byte-string"

    if "items" in definition:
        definition["items"] = process_node(definition["items"], root)

    if "properties" in definition:
        for key, value in definition["properties"].items():
            definition["properties"][key] = process_node(value, root)

    if "$defs" in definition:
        # Don't need these anymore
        definition.pop("$defs")

    return definition


def convert_anyof(anyof: list[dict], root) -> dict:
    nullable = False
    for option in anyof:
        option = process_node(option, root)
        if option["type"] == "null":
            # Replace this null option with nullable property
            nullable = True
            anyof.remove(option)
    if len(anyof) > 1:
        # TODO: Use sdfChoice
        raise NotImplementedError("Unions not supported yet")
    # Flatten
    definition = anyof[0]
    definition["nullable"] = nullable
    return definition


def convert_enum(definition: dict) -> dict:
    if len(definition["enum"]) == 1:
        # Probably means its a constant
        definition["const"] = definition["enum"][0]
        del definition["enum"]
    return definition


__all__ = ["from_json_schema"]



from typing import Any

from pydantic import BaseModel, ConfigDict, Field
from pydantic.alias_generators import to_camel


class CommonQualities(BaseModel):
    model_config = ConfigDict(extra="allow", alias_generator=to_camel)

    label: str | None = None
    description: str | None = None
    ref: str | None = Field(None, alias="sdfRef")

    def get_extra(self) -> dict[str, Any]:
        return self.__pydantic_extra__



import logging
from pathlib import Path
import subprocess
import tempfile

import pytest
from onedm import sdf


@pytest.fixture(scope="session")
def playground():
    with tempfile.TemporaryDirectory() as tmpdir:
        subprocess.run(
            ["git", "clone", "https://github.com/one-data-model/playground.git", tmpdir]
        )
        yield Path(tmpdir)


def test_parse_all_playground_models(playground: Path):
    IGNORED = [
        # Mixes numbers and integers
        "sdfobject-genericlevel.sdf.json",
    ]

    for path in playground.joinpath("sdfObject").glob("*.sdf.json"):
        if path.name in IGNORED:
            logging.warning("Skipping %s", path.name)
            continue

        loader = sdf.SDFLoader()
        logging.info("Parsing %s", path.name)
        loader.load_file(path)
        doc = loader.to_sdf()

        assert doc.data or doc.objects, f"No sdfData or sdfObject found in {path.name}"


def test_genericonoff(playground: Path):
    loader = sdf.SDFLoader()
    loader.load_file(playground / "sdfObject" / "sdfobject-genericonoff.sdf.json")
    doc = loader.to_sdf()

    assert doc.info.title == "Example Bluetooth mesh Generic OnOff Model"
    assert (
        doc.info.license
        == "https://github.com/one-data-model/oneDM/blob/master/LICENSE"
    )

    on_off = doc.objects["GenericOnOff"].properties["OnOff"]
    assert isinstance(on_off, sdf.AnyProperty)



from __future__ import annotations
from typing import Annotated, Literal, Union, Tuple

from pydantic import Field, NonNegativeInt

from .common import CommonQualities
from .data import (
    AnyData,
    ArrayData,
    BooleanData,
    Data,
    IntegerData,
    NumberData,
    ObjectData,
    StringData,
)


class NumberProperty(NumberData):
    observable: bool = True
    readable: bool = True
    writable: bool = True
    required: Tuple[Literal[True]] | None = Field(default=None, alias="sdfRequired")


class IntegerProperty(IntegerData):
    observable: bool = True
    readable: bool = True
    writable: bool = True
    required: Tuple[Literal[True]] | None = Field(default=None, alias="sdfRequired")


class BooleanProperty(BooleanData):
    observable: bool = True
    readable: bool = True
    writable: bool = True
    required: Tuple[Literal[True]] | None = Field(default=None, alias="sdfRequired")


class StringProperty(StringData):
    observable: bool = True
    readable: bool = True
    writable: bool = True
    required: Tuple[Literal[True]] | None = Field(default=None, alias="sdfRequired")


class ArrayProperty(ArrayData):
    observable: bool = True
    readable: bool = True
    writable: bool = True
    required: Tuple[Literal[True]] | None = Field(default=None, alias="sdfRequired")


class ObjectProperty(ObjectData):
    observable: bool = True
    readable: bool = True
    writable: bool = True
    required: Tuple[Literal[True]] | None = Field(default=None, alias="sdfRequired")


class AnyProperty(AnyData):
    observable: bool = True
    readable: bool = True
    writable: bool = True
    required: Tuple[Literal[True]] | None = Field(default=None, alias="sdfRequired")


Property = Union[
    Annotated[
        IntegerProperty
        | NumberProperty
        | BooleanProperty
        | StringProperty
        | ArrayProperty
        | ObjectProperty,
        Field(discriminator="type"),
    ],
    AnyProperty,
]


class Action(CommonQualities):
    input_data: Data | None = Field(None, alias="sdfInputData")
    output_data: Data | None = Field(None, alias="sdfOutputData")


class Event(CommonQualities):
    output_data: Data | None = Field(None, alias="sdfOutputData")


class Object(CommonQualities):
    properties: dict[str, Property] = Field(default_factory=dict, alias="sdfProperty")
    actions: dict[str, Action] = Field(default_factory=dict, alias="sdfAction")
    events: dict[str, Event] = Field(default_factory=dict, alias="sdfEvent")
    data: dict[str, Data] = Field(default_factory=dict, alias="sdfData")
    required: list[str] = Field(default_factory=list, alias="sdfRequired")
    # If array of objects
    min_items: NonNegativeInt | None = None
    max_items: NonNegativeInt | None = None


class Thing(CommonQualities):
    things: dict[str, Thing] = Field(default_factory=dict, alias="sdfThing")
    objects: dict[str, Object] = Field(default_factory=dict, alias="sdfObject")
    properties: dict[str, Property] = Field(default_factory=dict, alias="sdfProperty")
    actions: dict[str, Action] = Field(default_factory=dict, alias="sdfAction")
    events: dict[str, Event] = Field(default_factory=dict, alias="sdfEvent")
    data: dict[str, Data] = Field(default_factory=dict, alias="sdfData")
    required: list[str] = Field(default_factory=list, alias="sdfRequired")
    # If array of things
    min_items: NonNegativeInt | None = None
    max_items: NonNegativeInt | None = None



from onedm import sdf


def test_integer_data(test_model: sdf.SDF):
    assert isinstance(test_model.data["Integer"], sdf.IntegerData)
    assert test_model.data["Integer"].minimum == -2
    assert test_model.data["Integer"].maximum == 2
    assert test_model.data["Integer"].multiple_of == 2


def test_number_data(test_model: sdf.SDF):
    assert isinstance(test_model.data["Number"], sdf.NumberData)
    assert test_model.data["Number"].minimum == -1.5
    assert test_model.data["Number"].maximum == 1.5
    assert test_model.data["Number"].multiple_of == 0.5


def test_integer_choices(test_model: sdf.SDF):
    assert isinstance(test_model.data["Enum"], sdf.IntegerData)
    assert isinstance(test_model.data["Enum"].choices["One"], sdf.IntegerData)
    assert test_model.data["Enum"].choices["One"].const == 1


def test_boolean_data(test_model: sdf.SDF):
    assert isinstance(test_model.data["Boolean"], sdf.BooleanData)
    assert test_model.data["Boolean"].const == True


def test_string_data(test_model: sdf.SDF):
    assert isinstance(test_model.data["String"], sdf.StringData)
    assert test_model.data["String"].min_length == 10
    assert test_model.data["String"].max_length == 100
    assert test_model.data["String"].pattern == ".*"


def test_bytestring_data(test_model: sdf.SDF):
    assert isinstance(test_model.data["ByteString"], sdf.StringData)
    assert test_model.data["ByteString"].sdf_type == "byte-string"


def test_array_data(test_model: sdf.SDF):
    assert isinstance(test_model.data["Array"], sdf.ArrayData)
    assert test_model.data["Array"].min_items == 1
    assert test_model.data["Array"].max_items == 5

    assert isinstance(test_model.data["Array"].items, sdf.IntegerData)


def test_object_data(test_model: sdf.SDF):
    assert isinstance(test_model.data["Object"], sdf.ObjectData)

    assert isinstance(test_model.data["Object"].properties["prop1"], sdf.NumberData)
    assert test_model.data["Object"].required == ["prop1"]


def test_unknown_data(test_model: sdf.SDF):
    assert isinstance(test_model.data["Unknown"], sdf.AnyData)



from onedm import sdf


def test_integer_property(test_model: sdf.SDF):
    assert isinstance(test_model.properties["IntegerProperty"], sdf.IntegerProperty)
    assert test_model.properties["IntegerProperty"].label == "Example integer"


def test_number_property(test_model: sdf.SDF):
    assert isinstance(test_model.properties["NumberProperty"], sdf.NumberProperty)
    assert test_model.properties["NumberProperty"].label == "Example number"


def test_boolean_property(test_model: sdf.SDF):
    assert isinstance(test_model.properties["BooleanProperty"], sdf.BooleanProperty)
    assert test_model.properties["BooleanProperty"].label == "Example boolean"


def test_string_property(test_model: sdf.SDF):
    assert isinstance(test_model.properties["StringProperty"], sdf.StringProperty)
    assert test_model.properties["StringProperty"].label == "Example string"


def test_array_property(test_model: sdf.SDF):
    assert isinstance(test_model.properties["ArrayProperty"], sdf.ArrayProperty)
    assert test_model.properties["ArrayProperty"].label == "Example array"


def test_object_property(test_model: sdf.SDF):
    assert isinstance(test_model.properties["ObjectProperty"], sdf.ObjectProperty)
    assert test_model.properties["ObjectProperty"].label == "Example object"


def test_unknown_property(test_model: sdf.SDF):
    assert isinstance(test_model.properties["UnknownProperty"], sdf.AnyProperty)



"""Conversion from native types to sdfData."""

from enum import Enum
from typing import Type

from pydantic import TypeAdapter

from .data import Data, IntegerData, StringData
from .json_schema import from_json_schema

DataModel = TypeAdapter(Data)


def data_from_type(type_: Type) -> Data | None:
    """Create from a native Python or Pydantic type.

    None or null is not a supported type in SDF. In this case the return value
    will be None.
    """
    schema = TypeAdapter(type_).json_schema()

    if schema.get("type") == "null":
        # Null types not supported
        return None

    data = from_json_schema(schema)

    if "enum" in schema and issubclass(type_, Enum):
        data.choices = {}
        for member in type_:
            if isinstance(member.value, int):
                data.choices[member.name] = IntegerData(const=member.value)
            elif isinstance(member.value, str):
                data.choices[member.name] = StringData(const=member.value)
            else:
                raise TypeError("Unsupported enum type {type_}")
        data.enum = None

    return data



"""Data qualities.

Contains data qualities as defined by the standard, but divided into separate
types for correct validation and type hints.
"""

from __future__ import annotations

from abc import ABC
import datetime
from typing import Annotated, Any, Literal, Union

from pydantic import Field, NonNegativeInt
from pydantic_core import SchemaValidator, core_schema

from .common import CommonQualities


class DataQualities(CommonQualities, ABC):
    """Base class for all data qualities."""

    type: Literal["boolean", "number", "integer", "string", "object", "array"]
    sdf_type: str | None = Field(None, pattern=r"^[a-z][\-a-z0-9]*$")
    nullable: bool = True
    const: Any | None = None
    default: Any | None = None
    choices: dict[str, Data] | None = Field(None, alias="sdfChoice")

    def _get_base_schema(self) -> core_schema.CoreSchema:
        """Implemented by sub-classes."""
        raise NotImplementedError

    def get_pydantic_schema(self) -> core_schema.CoreSchema:
        """Get the Pydantic schema for this data quality."""
        if self.const is not None:
            schema = core_schema.literal_schema([self.const])
        elif self.choices is not None:
            schema = core_schema.union_schema(
                [
                    (choice.get_pydantic_schema(), name)
                    for name, choice in self.choices.items()
                ]
            )
        else:
            schema = self._get_base_schema()

        if self.default is not None:
            schema = core_schema.with_default_schema(schema, default=self.default)
        if self.nullable:
            schema = core_schema.nullable_schema(schema)
        return schema

    def validate_input(self, input: Any) -> Any:
        """Validate and coerce a value."""
        return SchemaValidator(self.get_pydantic_schema()).validate_python(input)


class NumberData(DataQualities):
    type: Literal["number"] = "number"
    unit: str | None = None
    minimum: float | None = None
    maximum: float | None = None
    exclusive_minimum: float | None = None
    exclusive_maximum: float | None = None
    multiple_of: float | None = None
    format: str | None = None
    const: float | None = None
    default: float | None = None

    def _get_base_schema(self) -> core_schema.FloatSchema | core_schema.DatetimeSchema:
        if self.sdf_type == "unix-time":
            return core_schema.datetime_schema(
                ge=(
                    datetime.datetime.fromtimestamp(self.minimum)
                    if self.minimum is not None
                    else None
                ),
                le=(
                    datetime.datetime.fromtimestamp(self.maximum)
                    if self.maximum is not None
                    else None
                ),
                gt=(
                    datetime.datetime.fromtimestamp(self.exclusive_minimum)
                    if self.exclusive_minimum is not None
                    else None
                ),
                lt=(
                    datetime.datetime.fromtimestamp(self.exclusive_maximum)
                    if self.exclusive_maximum is not None
                    else None
                ),
            )
        return core_schema.float_schema(
            ge=self.minimum,
            le=self.maximum,
            gt=self.exclusive_minimum,
            lt=self.exclusive_maximum,
            multiple_of=self.multiple_of,
        )

    def validate_input(self, input: Any) -> float:
        return super().validate_input(input)


class IntegerData(DataQualities):
    type: Literal["integer"] = "integer"
    unit: str | None = None
    minimum: int | None = None
    maximum: int | None = None
    exclusive_minimum: int | None = None
    exclusive_maximum: int | None = None
    multiple_of: int | None = None
    choices: dict[str, IntegerData] | None = Field(None, alias="sdfChoice")
    const: int | None = None
    default: int | None = None

    def _get_base_schema(self) -> core_schema.IntSchema:
        return core_schema.int_schema(
            ge=self.minimum,
            le=self.maximum,
            gt=self.exclusive_minimum,
            lt=self.exclusive_maximum,
            multiple_of=self.multiple_of,
        )

    def validate_input(self, input: Any) -> int:
        return super().validate_input(input)


class BooleanData(DataQualities):
    type: Literal["boolean"] = "boolean"
    const: bool | None = None
    default: bool | None = None

    def _get_base_schema(self) -> core_schema.BoolSchema:
        return core_schema.bool_schema()

    def validate_input(self, input: Any) -> bool:
        return super().validate_input(input)


class StringData(DataQualities):
    type: Literal["string"] = "string"
    enum: list[str] | None = None
    min_length: NonNegativeInt = 0
    max_length: NonNegativeInt | None = None
    pattern: str | None = None
    format: str | None = None
    content_format: str | None = None
    choices: dict[str, StringData] | None = Field(None, alias="sdfChoice")
    const: str | None = None
    default: str | None = None

    def _get_base_schema(
        self,
    ) -> core_schema.StringSchema | core_schema.BytesSchema | core_schema.LiteralSchema:
        if self.enum is not None:
            return core_schema.literal_schema(self.enum)
        if self.sdf_type == "byte-string":
            return core_schema.bytes_schema(
                min_length=self.min_length, max_length=self.max_length
            )
        if self.format == "uuid":
            return core_schema.uuid_schema()
        if self.format == "date-time":
            return core_schema.datetime_schema()
        if self.format == "date":
            return core_schema.date_schema()
        if self.format == "time":
            return core_schema.time_schema()
        if self.format == "uri":
            return core_schema.url_schema()
        return core_schema.str_schema(
            min_length=self.min_length,
            max_length=self.max_length,
            pattern=self.pattern,
        )

    def validate_input(self, input: Any) -> str | bytes:
        return super().validate_input(input)


class ArrayData(DataQualities):
    type: Literal["array"] = "array"
    min_items: NonNegativeInt = 0
    max_items: NonNegativeInt | None = None
    unique_items: bool = False
    items: Data | None = None
    const: list | None = None
    default: list | None = None

    def _get_base_schema(self) -> core_schema.ListSchema | core_schema.SetSchema:
        if self.unique_items:
            return core_schema.set_schema(
                self.items.get_pydantic_schema(),
                min_length=self.min_items,
                max_length=self.max_items,
            )
        return core_schema.list_schema(
            self.items.get_pydantic_schema(),
            min_length=self.min_items,
            max_length=self.max_items,
        )

    def validate_input(self, input: Any) -> list | set:
        return super().validate_input(input)


class ObjectData(DataQualities):
    type: Literal["object"] = "object"
    required: list[str] = Field(default_factory=list)
    properties: dict[str, Data] | None = None
    const: dict[str, Any] | None = None
    default: dict[str, Any] | None = None

    def _get_base_schema(self) -> core_schema.TypedDictSchema:
        required = self.required or []
        fields = {
            name: core_schema.typed_dict_field(
                property.get_pydantic_schema(), required=name in required
            )
            for name, property in self.properties.items()
        }
        return core_schema.typed_dict_schema(fields)

    def validate_input(self, input: Any) -> dict:
        return super().validate_input(input)


class AnyData(DataQualities):
    type: Literal[None] = None

    def _get_base_schema(self) -> core_schema.AnySchema:
        return core_schema.any_schema()


Data = Union[
    Annotated[
        IntegerData | NumberData | BooleanData | StringData | ObjectData | ArrayData,
        Field(discriminator="type"),
    ],
    AnyData,
]

ObjectData.model_rebuild()
ArrayData.model_rebuild()



from datetime import datetime

from pydantic import BaseModel, Field

from .definitions import Data, Event, Object, Property, Thing


class Information(BaseModel):
    title: str | None = None
    version: str | None = None
    copyright: str | None = None
    license: str | None = None
    features: list[str] = Field(default_factory=list)
    modified: datetime | None = None


class SDF(BaseModel):
    info: Information = Field(default_factory=Information)
    namespace: dict[str, str] = Field(default_factory=dict)
    default_namespace: str | None = Field(None, alias="defaultNamespace")
    things: dict[str, Thing] = Field(default_factory=dict, alias="sdfThing")
    objects: dict[str, Object] = Field(default_factory=dict, alias="sdfObject")
    properties: dict[str, Property] = Field(default_factory=dict, alias="sdfProperty")
    events: dict[str, Event] = Field(default_factory=dict, alias="sdfEvent")
    data: dict[str, Data] = Field(default_factory=dict, alias="sdfData")

    def to_json(self) -> str:
        return self.model_dump_json(indent=2, exclude_defaults=True, by_alias=True)
